﻿module ParasitemiaCore.Config

open System

open Const
open UnitsOfMeasure

type Debug =
    | DebugOff
    | DebugOn of string // Output directory.

type Parameters =
    {
        rbcDiameter : float<μm>
        resolution : float<ppi>

        ratioAreaPaleCenter : float32 // The area of the second opening is 'ratioSecondAreaOpen' * mean RBC area. It's applied only if greater than 'initialAreaOpen'.

        granulometryRange : float32 // The radius will be seeked from radius - granulometryRange * radius to radius + granulometryRange * radius.

        minRbcRadius : float32 // Factor of the mean RBC radius.
        maxRbcRadius : float32 // Factor of the mean RBC radius.

        LPFStandardDeviationParasite : float<μm> // Sigma parameter of the gaussian to remove the high frequency noise.
        LPFStandardDeviationRBC : float<μm>

        // Ellipse.
        nbPickElementsMin : int
        factorNbValidPick : float // The number of computed ellipse per edge pixel.
        factorNbMaxPick : float

        // Parasites detection.
        darkStainLevel : float // Lower -> more sensitive. Careful about illumination on the borders.
        maxDarkStainRatio : float // When a cell must own less than this ratio to be a RBC.

        parasiteRadiusRatio : float32 // The ratio of the parasite radius of the RBC radius.
        minimumParasiteAreaRatio : float32 // Factor of a RBC area. 0.5 means the half of RBC area.

        cytoplasmSizeRatio : float32
        cytoplasmSensitivity : float // between 0 (the least sensitive) and 1 (the most sensitive).

        nucleusAreaRatio : float32 // Factor of a RBC area. 0.5 means the half of RBC area.
        infectionSensitivity : float // between 0 (the least sensitive) and 1 (the most sensitive).

        // [<Obsolete>] standardDeviationMaxRatio: float // The standard deviation of the pixel values of a cell can't be greater than standardDeviationMaxRatio * global standard deviation
        minimumCellAreaFactor : float32 // Factor of the mean RBC area. A cell with an area below this will be rejected.
    }

let defaultParameters =
    {
        rbcDiameter = 7.5<μm>
        resolution = 230.e3<ppi> // 230.e3<ppi> Correspond to 50X.

        ratioAreaPaleCenter = 2.f / 5.f // The ratio between an RBC area and the area of the its pale center.

        granulometryRange = 0.5f

        minRbcRadius = -0.23f
        maxRbcRadius = 0.23f

        LPFStandardDeviationParasite = 0.15<μm>
        LPFStandardDeviationRBC = 0.22<μm>

        nbPickElementsMin = 10
        factorNbValidPick = 0.06 //1.0
        factorNbMaxPick = 4.

        darkStainLevel = 1.
        maxDarkStainRatio = 0.1 // 10 %

        parasiteRadiusRatio = 0.5f // 50 %
        minimumParasiteAreaRatio = 0.02f // 2 %

        cytoplasmSizeRatio = 1.f / 5.f
        cytoplasmSensitivity = 0.96

        nucleusAreaRatio = 0.01f // 1.0 %
        infectionSensitivity = 0.92

        // standardDeviationMaxRatio = 0.6 // Obsolete.
        minimumCellAreaFactor = 0.4f
    }

type RBCRadius (radius : float32, parameters : Parameters) =
    member this.Pixel = radius
    member this.μm : float<μm> =
        1.<px> * (float radius) / parameters.resolution |> inchToμm

    member this.Min = radius + parameters.minRbcRadius * radius
    member this.Max = radius + parameters.maxRbcRadius * radius

    member this.Area = PI * radius ** 2.f
    member this.MinArea = parameters.minimumCellAreaFactor * this.Area

    member this.ParasiteRadius = parameters.parasiteRadiusRatio * radius
    member this.CytoplasmSize = parameters.cytoplasmSizeRatio * radius

    member this.NucleusArea = parameters.nucleusAreaRatio * this.Area
    member this.MinimumParasiteArea = parameters.minimumParasiteAreaRatio * this.Area

    override this.ToString() =
        sprintf "%d px (%.1f μm)" (Utils.roundInt <| 2.f * radius) (2. * this.μm)

type Config (param : Parameters) =
    let RBCadiusInPixels (rbcDiameter : float<μm>) (resolution : float<ppi>) : float32 =
        let rbcRadiusInch : float<inch> = (μmToInch rbcDiameter) / 2.
        let rbcRadiusPx : float<px> = resolution * rbcRadiusInch
        float32 rbcRadiusPx

    let mutable parameters : Parameters = param
    let mutable rbcRadiusByResolution = RBCRadius(RBCadiusInPixels parameters.rbcDiameter parameters.resolution, parameters)
    let mutable rbcRadius = RBCRadius(0.f, parameters)

    new () = Config(defaultParameters)

    member this.Parameters
        with get() = parameters
        and set(param) =
            parameters <- param
            rbcRadiusByResolution <- RBCRadius(RBCadiusInPixels parameters.rbcDiameter parameters.resolution, param)
            rbcRadius <- RBCRadius(rbcRadius.Pixel, param)

    member val Debug = DebugOff with get, set

    member this.LPFStandardDeviationParasite =
        let stdDeviation : float<px> = (μmToInch parameters.LPFStandardDeviationParasite) * parameters.resolution
        float stdDeviation

    member this.LPFStandardDeviationRBC =
        let stdDeviation : float<px> = (μmToInch parameters.LPFStandardDeviationRBC) * parameters.resolution
        float stdDeviation

    member this.RBCRadiusByResolution = rbcRadiusByResolution
    member this.RBCRadius = rbcRadius

    member this.SetRBCRadius (radiusPixel : float32) =
        rbcRadius <- RBCRadius(radiusPixel, parameters)

    member this.Copy () =
        this.MemberwiseClone() :?> Config

