﻿module ParasitemiaCore.Ellipse

open System
open System.Collections.Generic
open System.Drawing

open MathNet.Numerics.LinearAlgebra

open Emgu.CV
open Emgu.CV.Structure

open Utils
open Config
open MatchingEllipses
open Const

// This is a ratio of the biggest radius.
let minimumDistanceBetweenDrawnPoints = 0.6

/// <summary>
/// Try to build an ellipse from three points and two tangents passing by the first and the second point.
/// 'Ellipse.A' is always equal or greater than Ellipse.B.
/// 'Ellipse.Alpha' is between 0 and Pi.
/// </summary>
let ellipse (p1x : float) (p1y : float) (m1 : float) (p2x : float) (p2y : float) (m2 : float) (p3x : float) (p3y : float) : Types.Ellipse option =
    let p0 = pointFromTwoLines (Types.Line(float32 m1, float32 (p1y - m1 * p1x))) (Types.Line(float32 m2, float32(p2y - m2 * p2x)))
    let p0x, p0y = float p0.X, float p0.Y

    let s =
        matrix
            [
                [ 1.;   0.;  0.  ]
                [ 0.;   0.; -0.5 ]
                [ 0.; -0.5;  0.  ]
            ]

    let v0 = matrix [[ 1.; p0x; p0y ]]
    let v1 = matrix [[ 1.; p1x; p1y ]]
    let v2 = matrix [[ 1.; p2x; p2y ]]
    let v3 = matrix [[ 1.; p3x; p3y ]]

    let p = (v3.Stack(v1).Stack(v2).Determinant() * v0).Stack(v0.Stack(v3).Stack(v2).Determinant() * v1).Stack(v0.Stack(v1).Stack(v3).Determinant() * v2).Transpose()
    let conicMat = p * s.Inverse() * p.Transpose()
    let a = conicMat.[0, 0]
    let b = conicMat.[0, 1]
    let c = conicMat.[1, 1]
    let d = conicMat.[0, 2]
    let e = conicMat.[1, 2]
    let f = conicMat.[2, 2]

    // Center.
    let cx = b / a
    let cy = d / a

    let at = c * f - e ** 2. + (e * d - b * f) * cx + (b * e - c * d) * cy
    if at = 0. then
        None
    else
        let q = (-1. / at) * (matrix [[ a * f - d ** 2.0; b * d - a * e ]; [ b * d - a * e; a * c - b ** 2.0 ]])
        let eigen = q.Evd()
        let eigenValues = eigen.EigenValues
        let lambda = eigenValues.[1].Real
        let mu = eigenValues.[0].Real

        if lambda <= 0. || mu <= 0. then
            None
        else
            let r1, r2 = 1. / (sqrt lambda), 1. / (sqrt mu)

            let eigenVectors = eigen.EigenVectors
            let v_a = eigenVectors.[0, 0]
            let v_b = eigenVectors.[1, 0] // [0, 1]

            // Angle against the longest axis.
            let phi = (if r2 > r1 then atan (v_b / v_a) else atan (v_a / v_b))

            let phi' = if phi < 0. then phi + Math.PI else phi
            let majorAxis, minorAxis = if r1 > r2 then r1, r2 else r2, r1

            Some (Types.Ellipse(float32 cx, float32 cy, float32 majorAxis, float32 minorAxis, float32 phi'))

let inline private vectorRotation (px : float32) (py : float32) (vx : float32) (vy : float32) (p0x : float32) (p0y : float32) : float32 =
    if py > p0y then
        if vx > 0.f then -1.f else 1.f
    elif py < p0y then
        if vx < 0.f then -1.f else 1.f
    elif px > p0x then
        if vy < 0.f then -1.f else 1.f
    else // p1x < px
        if vy > 0.f then -1.f else 1.f

let private areVectorsValid (p1x : float32) (p1y : float32) (p2x : float32) (p2y : float32) (v1x : float32) (v1y : float32) (v2x : float32) (v2y : float32) : (float32 * float32) option =
    let m1 = -v1x / v1y
    let m2 = -v2x / v2y

    let b1 = -m1 * p1x + p1y
    let b2 = -m2 * p2x + p2y
    let p0x = -(b1 - b2) / (m1 - m2)
    let p0y = -(m2 * b1 - m1 * b2) / (m1 - m2)

    let rot1 = vectorRotation p1x p1y v1x v1y p0x p0y
    let rot2 = vectorRotation p2x p2y v2x v2y p0x p0y

    if rot1 = rot2 then
        None
    else
        let alpha1 =
            let alpha1' = atan2 (p1y - p0y) (p1x - p0x)
            if alpha1' < 0.f then 2.f * PI + alpha1' else alpha1'

        let alpha2 =
            let alpha2' = atan2 (p2y - p0y) (p2x - p0x)
            if alpha2' < 0.f then 2.f * PI + alpha2' else alpha2'

        let diff = rot1 * alpha1 + rot2 * alpha2

        if diff > PI || (diff < 0.f && diff > -PI) then
            Some (m1, m2)
        else
            None

/// <summary>
/// Build a set of ellipses as a 'MatchingEllipses' object by finding ellipses with the given edges and gradient.
/// </summary>
let find (edges : Matrix<byte>)
         (xGradient : Matrix<float32>)
         (yGradient : Matrix<float32>)
         (config : Config) : MatchingEllipses =

    let r1, r2 = config.RBCRadius.Min, config.RBCRadius.Max
    let incrementWindowDivisor = 4.f

    // We choose a window size for which the biggest ellipse can always be fitted in.
    let windowSize = roundInt (2.f * r2)
    let factorNbValidPick = config.Parameters.factorNbValidPick
    let factorNbMaxPick = config.Parameters.factorNbMaxPick
    let nbPickElementsMin = config.Parameters.nbPickElementsMin

    let increment = windowSize / (int incrementWindowDivisor)

    let radiusTolerance = (r2 - r1) * 0.2f

    let squaredMinimumDistance = (float config.RBCRadius.Pixel * minimumDistanceBetweenDrawnPoints) ** 2.
    let inline squaredDistance x1 y1 x2 y2 = (x1 - x2) ** 2. + (y1 - y2) ** 2.

    let h = edges.Height
    let w = edges.Width
    let h_f = float32 h
    let w_f = float32 w

    let mutable last_i, last_j = Int32.MaxValue, Int32.MaxValue

    let currentElements = List<Point>()

    let edgesData = edges.Data
    let xDirData = xGradient.Data
    let yDirData = yGradient.Data

    let rng = Random(42)

    let ellipses = MatchingEllipses(config.RBCRadius.Pixel)

    for window_i in -windowSize + increment .. increment .. h - increment do
        for window_j in -windowSize + increment .. increment .. w - increment do

            let window_i_begin = if window_i < 0 then 0 else window_i
            let window_i_end = if window_i + windowSize - 1 >= h then h - 1 else window_i + windowSize - 1
            let window_j_begin = if window_j < 0 then 0 else window_j
            let window_j_end = if window_j + windowSize - 1 >= w then w - 1 else window_j + windowSize - 1

            // Remove old elements.
            let indexFirstElement = currentElements.FindIndex(fun p -> p.X >= window_j_begin)
            if indexFirstElement > 0 then
                currentElements.RemoveRange(0, indexFirstElement)

            // Add the new elements.
            let newElemsBegin_j = window_j + windowSize - increment
            let newElemsEnd_j = window_j + windowSize - 1
            for j = (if newElemsBegin_j < 0 then 0 else newElemsBegin_j) to (if newElemsEnd_j >= w then w - 1 else newElemsEnd_j) do
                for i = window_i_begin to window_i_end do
                    if edgesData.[i, j] = 1uy then
                        currentElements.Add(Point(j, i))

            if currentElements.Count >= nbPickElementsMin then
                let mutable nbOfPicks = (float currentElements.Count) * factorNbMaxPick |> int
                let mutable nbOfValidPicks = (float currentElements.Count) * factorNbValidPick |> int
                while nbOfPicks > 0 && nbOfValidPicks > 0 do
                    let p1 = currentElements.[rng.Next(currentElements.Count)]
                    let p2 = currentElements.[rng.Next(currentElements.Count)]
                    let p3 = currentElements.[rng.Next(currentElements.Count)]
                    if p1 <> p2 && p1 <> p3 && p2 <> p3 then
                        nbOfPicks <- nbOfPicks - 1
                        let p1yf, p1xf = float p1.Y, float p1.X
                        let p2yf, p2xf = float p2.Y, float p2.X
                        let p3yf, p3xf = float p3.Y, float p3.X
                        if squaredDistance p1xf p1yf p2xf p2yf >= squaredMinimumDistance &&
                           squaredDistance p1xf p1yf p3xf p3yf >= squaredMinimumDistance &&
                           squaredDistance p2xf p2yf p3xf p3yf >= squaredMinimumDistance
                        then
                            match areVectorsValid (float32 p1xf) (float32 p1yf) (float32 p2xf) (float32 p2yf) -xDirData.[p1.Y, p1.X] -yDirData.[p1.Y, p1.X] -xDirData.[p2.Y, p2.X] -yDirData.[p2.Y, p2.X] with
                            | Some (m1, m2) ->
                                match ellipse p1xf p1yf (float m1) p2xf p2yf (float m2) p3xf p3yf with
                                | Some e when e.Cx > 0.f && e.Cx <  w_f - 1.f && e.Cy > 0.f && e.Cy < h_f - 1.f &&
                                              e.A >= r1 - radiusTolerance && e.A <= r2 + radiusTolerance && e.B >= r1 - radiusTolerance && e.B <= r2 + radiusTolerance ->
                                    nbOfValidPicks <- nbOfValidPicks - 1
                                    ellipses.Add e
                                | _ -> ()
                            | _ -> ()

        currentElements.Clear()

    ellipses

