﻿module ParasitemiaUI.Main

open System
open System.IO
open System.Threading

open Emgu.CV
open Emgu.CV.Structure

open Logger

open ParasitemiaCore.Utils
open ParasitemiaCore.Config

type Input =
    | File of string
    | Dir of string

type RunningMode =
    | CmdLine of Input * string // A file or a directory to process and the output directory.
    | Window of string option // An optional path to a file to open can be given in window mode.

type Arguments = RunningMode * bool

let parseArgs (args : string[]) : Arguments =

    let output = Array.tryFindIndex ((=) "--output") args

    let runningMode =
        match Array.tryFindIndex ((=) "--folder") args, output with
        | Some i, Some i_output when i < args.Length - 2 && i_output < args.Length - 2 ->
            CmdLine ((Dir args.[i + 1]), args.[i_output + 1])
        | _ ->
            match Array.tryFindIndex ((=) "--file") args, output with
            | Some i, Some i_output when i < args.Length - 2 && i_output < args.Length - 2 ->
                CmdLine ((File args.[i + 1]), args.[i_output + 1])
            |_ ->
                Window (if args.Length > 0 && not (args.[0].StartsWith("--")) then Some args.[0] else None)

    runningMode, Array.exists ((=) "--debug") args

let showArgsHelp () =
    printfn "Usage of Parasitemia :"
    printfn "Non-interactive mode:"
    printfn "  %s (--folder <folder>|--file <file>) --output <folder> [--debug]" System.AppDomain.CurrentDomain.FriendlyName
    printfn "    --folder <folder> : an input folder containing images to analyze"
    printfn "    --file <file> : an image file to be analyzed"
    printfn "    --output <folder> : a folder to put the results"
    printfn "    --debug : output more information like intermediate images if set"

    printfn "Interactive mode:"
    printfn "  %s [<document-file>] [--debug]" System.AppDomain.CurrentDomain.FriendlyName
    printfn "    <document-file> : a PIAZ file to automatically open at startup"
    printfn "    --debug : output information like intermediate images if set in the current directory"

[<System.Runtime.InteropServices.DllImport("kernel32.dll")>]
extern bool AttachConsole(int dwProcessId)

[<EntryPoint>]
[<STAThread()>]
let main args =
    // To redirect stdout to the attached console.
    AttachConsole(-1) |> ignore // -1 to attach to the parent process.

    if Array.exists (fun e -> e = "--help" || e = "-h") args then
        showArgsHelp ()
        0
    else
        try
            Log.Info "Starting of Parasitemia UI ..."

            let result =
                match parseArgs args with
                | mode, debug ->
                    let config = Config(defaultParameters)

                    match mode with
                    | CmdLine (input, output) ->
                        if debug then
                            config.Debug <- DebugOn output

                        Directory.CreateDirectory output |> ignore

                        use logFile = new StreamWriter(new FileStream(Path.Combine(output, "log.txt"), FileMode.Append, FileAccess.Write))
                        let listener = { new IListener with member this.NewEntry severity header mess = logFile.WriteLine(header + " : " + mess) }
                        Log.AddListener(listener)

                        Log.Info "=== New run : %O %s ===" DateTime.Now  (if debug then "[DEBUG]" else "[RELEASE]")

                        let files = match input with
                                    | File file -> [ file ]
                                    | Dir dir -> Directory.EnumerateFiles dir |> List.ofSeq

                        use resultFile = new StreamWriter(new FileStream(Path.Combine(output, "results.txt"), FileMode.Append, FileAccess.Write))

                        let images = [ for file in files -> Path.GetFileNameWithoutExtension(FileInfo(file).Name), config.Copy(), new Image<Bgr, byte>(file) ]

                        Log.LogWithTime Severity.INFO (fun () ->
                            match ParasitemiaCore.Analysis.doMultipleAnalysis images None with
                            | Some results ->
                                for id, cells in results do
                                    let config, img = images |> List.pick (fun (id', config', img') -> if id' = id then Some (config', img') else None)
                                    img.Dispose()
                                    let total, infected = countCells cells
                                    fprintf resultFile "File: %s %d %d %.2f (diameter: %O)\n" id total infected (100. * (float infected) / (float total)) config.RBCRadius
                            | None ->
                                fprintf resultFile "Analysis aborted"
                            Some ()) "Whole analyze" |> ignore

                        Log.RmListener(listener)
                        0

                    | Window fileToOpen ->
                        if debug then config.Debug <- DebugOn "."
                        GUI.run config fileToOpen

            Log.Info "Parasitemia UI closed"
            result

        with
        | ex ->
            Log.Fatal "Error: %A" ex
            1